import multiprocessing
import re
from functools import partial

import numpy as np
import pandas as pd
import scipy.stats as stats
import statsmodels.api as sm
from scipy.stats import pearsonr
from tqdm import tqdm


def calc_corr(
    *,
    first_file,
    second_file,
    samples_col,
    output_file,
):
    """
    Calculate the pearson's correlation between same genes in two scoring matices.

    Parameters
    ----------
    first_file : str
        the path to the first scores file.
    second_file : str
        the path to the second scores file.
    samples_col : str
        the column containing the samples IDs.
    output_file : str
        the path to the output file with correlation values.

    Returns
    -------
    pd.DataFrame
        the correlation dataframe.

    """
    with open(first_file) as f:
        genes_01 = re.split('\s+', f.readline().strip('\n'))
        genes_01.remove(samples_col)
    with open(second_file) as f:
        genes_02 = re.split('\s+', f.readline().strip('\n'))
        genes_02.remove(samples_col)
    as_set = set(genes_01)
    common_genes = as_set.intersection(genes_02)
    genes = list(common_genes)
    corr_info = []
    first_df = pd.read_csv(first_file, sep='\t', index_col=False)
    second_df = pd.read_csv(second_file, sep='\t', index_col=False)
    for gene in tqdm(genes, desc='calculating correlation'):
        gene_df = pd.merge(first_df[[samples_col, gene]], second_df[[samples_col, gene]], on=samples_col)
        gene_df.replace([np.inf, -np.inf, np.nan], 0.0, inplace=True)
        corr, pval = pearsonr(gene_df[gene + '_x'], gene_df[gene + '_y'])
        corr_info.append([gene, corr, pval])
    corr_df = pd.DataFrame(corr_info, columns=['target', 'corr', 'p_value']).sort_values(by=['p_value'])
    corr_df.to_csv(output_file, sep='\t', index=False)
    return corr_df


def run_linear(gene_col, x_set, y_set):
    """
    Helper function to run linear regression association.

    Parameters
    ----------
    gene_col : tuple
        a tuple from df.iteritems()
    x_set : pd.DataFrame
        a dataframe with the covariates.
    y_set : pd.Series
        the target.

    Returns
    -------
    List
         a list with gene name, pvalues, coefs and std err.

    """
    x_set[gene_col[0]] = gene_col[1]
    x_set = sm.add_constant(x_set)
    linear_model = sm.OLS(y_set, x_set)
    result = linear_model.fit()
    pval = list(result.pvalues)
    beta_coef = list(result.params)[-1]
    std_err = result.bse[-1]
    return [gene_col[0]] + pval + [beta_coef, std_err]


def run_logit(gene_col, x_set, y_set):
    """
    Helper function to run logistic regression association.

    Parameters
    ----------
    gene_col : tuple
        a tuple from df.iteritems()
    x_set : pd.DataFrame
        a dataframe with the covariates.
    y_set : pd.Series
        the target.

    Returns
    -------
    List
         a list with gene name, pvalues, coefs and std err.

    """
    try:
        x_set[gene_col[0]] = gene_col[1]
        x_set = sm.add_constant(x_set)
        logit_model = sm.Logit(y_set, x_set)
        result = logit_model.fit_regularized()
        pval = list(result.pvalues)
        std_err = result.bse[-1]
        coef = result.params[-1]
    except:
        return
    return [gene_col[0]] + pval + [coef, std_err]


def run_mannwhitneyu(*, df, genes, cases_column, **kwargs):
    """
    Helper function to run mann whitney u association.

    Parameters
    ----------
    df : df.DataFrame
        the dataframe with all info.
    genes : List
        the list of genes to run the association
    cases_column : str
        the taget column
    kwargs

    Returns
    -------
    pd.DataFrame
        dataframe containing the p_values.

    """
    p_values = []
    df_by_cases = df.groupby(cases_column)
    kwargs['logger'].info(df_by_cases.size())
    if kwargs['cases'] and kwargs['controls']:
        cc = [kwargs['cases'], kwargs['controls']]
    else:
        cc = list(df_by_cases.groups.keys())
    if len(cc) > 2:
        Warning('There are more than two categories here. We will only consider the first two categories.')
    for gene in tqdm(genes, desc='Calculating p_values for genes'):
        case_0 = df_by_cases.get_group(cc[0])[gene].tolist()
        case_1 = df_by_cases.get_group(cc[1])[gene].tolist()
        try:
            u_statistic, p_val = stats.mannwhitneyu(case_0, case_1)
        except:
            continue
        p_values.append([gene, u_statistic, p_val])
    cols = ['genes', 'statistic', 'p_value']
    p_values_df = pd.DataFrame(p_values, columns=cols).sort_values(by=['p_value'])
    return p_values_df


def run_ttest(*, df, genes, cases_column, **kwargs):
    p_values = []
    df_by_cases = df.groupby(cases_column)
    kwargs['logger'].info(df_by_cases.size())
    if kwargs['cases'] and kwargs['controls']:
        cc = [kwargs['cases'], kwargs['controls']]
    else:
        cc = list(df_by_cases.groups.keys())
    if len(cc) > 2:
        Warning('There are more than two categories here. We will only consider the first two categories.')

    for gene in tqdm(genes, desc='Calculating p_values for genes'):
        case_0 = df_by_cases.get_group(cc[0])[gene].tolist()
        case_1 = df_by_cases.get_group(cc[1])[gene].tolist()
        try:
            statistic, p_val = stats.ttest_ind(case_0, case_1)
        except:
            continue
        p_values.append([gene, statistic, p_val])
    cols = ['genes', 'statistic', 'p_value']
    p_values_df = pd.DataFrame(p_values, columns=cols).sort_values(by=['p_value'])
    return p_values_df


def get_pvals_logit(*, df, genes, cases_column, **kwargs):
    covariates = kwargs['covariates']
    df[cases_column] = np.interp(
        df[cases_column], (df[cases_column].min(), df[cases_column].max()), (0, 1))
    kwargs['logger'].info(df.groupby(cases_column).size())
    y_set = df[[cases_column]]
    x_set = df[covariates]
    genes_df = df[genes]
    pool = multiprocessing.Pool(processes=kwargs['processes'])
    partial_func = partial(run_logit, x_set=x_set, y_set=y_set)
    p_values = list(pool.imap(partial_func, genes_df.iteritems()))
    cols = ['genes', 'const_pval'] + covariates + ['p_value', 'coef', 'std_err']
    p_values_df = pd.DataFrame(list(filter(None, p_values)), columns=cols).sort_values(by=['p_value'])
    return p_values_df


def get_pvals_linear(*, df, genes, cases_column, **kwargs):
    covariates = kwargs.get('covariates')
    y_set = df[[cases_column]]
    x_set = df[covariates]
    genes_df = df[genes]
    pool = multiprocessing.Pool(processes=kwargs['processes'])
    partial_func = partial(run_linear, x_set=x_set, y_set=y_set)
    p_values = list(pool.imap(partial_func, genes_df.iteritems()))
    cols = ['genes', 'const_pval'] + covariates + ['p_value', 'beta_coef', 'std_err']
    p_values_df = pd.DataFrame(p_values, columns=cols).sort_values(by=['p_value'])
    return p_values_df



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from adjustText import adjust_text
from qmplot import qqplot
from tqdm import tqdm
from pybiomart import Dataset
from textwrap import wrap



def gene_length_normalize(*, genes_info, genes_col='HGNC symbol', length_col='gene_length', scores_df, samples_col):
    """
    Normalize dataset by gene length. if gene lengths file is not provided, info will be retrieved from ensembl.

    Parameters
    ----------
    genes_info : str
        file containing gene lengths. If file is not provided, info will be retrieved from ensembl
    genes_col : str
        column containing gene names.
    length_col : str
        column containing the length of each gene.
    scores_df : pd.DataFrame
        dataframe containing data to normalize.
    samples_col : str
        column containing samples IDs.

    Returns
    -------
    pd.Dataframe
        dataframe containing normalized dataframe.

    """
    unnormalized = []
    if not genes_info:
        dataset = Dataset(name='hsapiens_gene_ensembl',
                          host='http://www.ensembl.org')
        genes_df = dataset.query(attributes=['hgnc_symbol', 'start_position', 'end_position'])
        genes_df['gene_length'] = genes_df['Gene end (bp)'] - genes_df['Gene start (bp)']
    else:
        genes_df = pd.read_csv(genes_info, sep='\t')
    genes_lengths = genes_df.set_index(genes_col).to_dict()[length_col]
    for (name, data) in tqdm(scores_df.drop(columns=[samples_col]).iteritems(), desc="Normalizing genes scores"):
        if name not in genes_lengths.keys():
            unnormalized.append(name)
            continue
        # normalize genes by length
        scores_df[name] = round(scores_df[name] / genes_lengths[name], 5)
    scores_df = scores_df.drop(unnormalized, axis=1)
    return scores_df


def draw_manhattan(*, data, chr_col, pos_col, pvals_col, genes_col, manhattan_output):
    """
    Generate  manhattan plot from a given dataset.

    Parameters
    ----------
    data : pd.DataFrame
        a dataframe with pvalues and gene information.
    chr_col : str
        the column with the chromosomes.
    pos_col : str
        the column containing the position/start.
    pvals_col : str
        the column containing the position/start.
    genes_col : str
        the column containing gene names.
    manhattan_output : str
        the path to output the manhattan plot image.

    Returns
    -------
    Manhattan plot

    """
    data.drop_duplicates(subset=[genes_col], inplace=True)
    data.dropna(subset=[pvals_col], inplace=True)
    data['-logp'] = - np.log10(data[pvals_col])
    data = data.dropna(subset=[genes_col, chr_col, pos_col], axis=0)
    data[chr_col].replace('X', 23, inplace=True)
    data[chr_col] = data[chr_col].astype('int64')
    data = data.sort_values([chr_col, pos_col])
    data.reset_index(inplace=True, drop=True)
    data['ind'] = data.index
    df_grouped = data.groupby((chr_col))
    fig = plt.figure(figsize=(14, 8))  # Set the figure size
    ax = fig.add_subplot(111)
    colors = ['black', 'grey']
    for num, (name, group) in enumerate(df_grouped):
        group.plot(kind='scatter', x='ind', y='-logp', color=colors[num % len(colors)], ax=ax)
    ax.set_ylim(0.0, max(data["-logp"]) + 1)

    for i in data[data['-logp'] > 5.6].index:
        ax.annotate(xy=(data.ind[i], data['-logp'][i]),
                              xytext=(data.ind[i] + 0.2, data['-logp'][i] + 0.2),
                              text=data[genes_col][i],
                              arrowprops=dict(arrowstyle="->", color='b', lw=0.5),
                              horizontalalignment='left', size='medium', rotation=20, color='black')
    # adjust_text(anno, arrowprops=dict(arrowstyle="->", color='b', lw=0.5))
    chrom_df = data.groupby(chr_col)['ind'].median()
    data['color group'] = data[chr_col].apply(lambda x: 'A' if x % 2 == 0 else 'B')
    fig.suptitle(manhattan_output.split('.')[0])
    ax.set_ylim(0.0, max(data["-logp"]) + 1)
    ax.set_xlabel(chr_col)
    ax.set_xticks(chrom_df)
    ax.set_xticklabels(chrom_df.index)
    ax.axhline(-np.log10(1.00e-05), c='blue', ls='--')
    ax.axhline(-np.log10(5.00e-08), c='red', ls='--')
    fig.savefig(manhattan_output)
    return ax


def draw_qqplot(*, pvals, qq_output):
    """
    Generate QQ-plot for given data.

    Parameters
    ----------
    pvals : pd.Series
        the list of p_values.
    qq_output : str
        the path to output the QQplot image.

    Returns
    -------
    QQPlot

    """
    pvals.dropna(inplace=True)
    f, ax = plt.subplots(figsize=(6, 6), facecolor="w", edgecolor="k")
    title = "\n".join(wrap(qq_output.split('.')[0], 50))
    qqplot(data=pvals,
           title=title,
           marker="o",
           xlabel=r"Expected $-log_{10}{(P)}$",
           ylabel=r"Observed $-log_{10}{(P)}$",
           ax=ax)
    plt.savefig(qq_output)
    return ax


def merge_files(*, files_lst, sep, by, cols=None):
    """
    Merge a list of files with the same format into one dataframe.

    Parameters
    ----------
    files_lst
        a list with files to merge.
    sep
        the column seperator.
    by
        the common column between all files.
    cols
        selected columns to remain in dataframe.

    Returns
    -------
    merged dataframe

    """
    df = pd.read_csv(files_lst[0], sep=sep)
    for file in files_lst[1:]:
        df2 = pd.read_csv(file, sep=sep)
        df = pd.merge(df, df2, on=by)
    if cols:
        df = df[[cols]]
    return df




import joblib
import numpy as np
import pycaret.classification as pycl
import pycaret.regression as pyreg
import sklearn.metrics as metrics

from .helpers import write_output, generate_confusion_matrix, generate_scatterplot


def regression_model(
    *,
    y_col,
    training_set,
    normalize,
    folds,
    metric,
    model_name,
    testing_set,
    imbalanced,
    seed,
    include_models,
    normalize_method,
    feature_selection
):
    """
    Build a regression model for prediction.

    Parameters
    ----------
    y_col : str
        the name of the target column.
    training_set : pd.DataFrame
        DataFrame containing the training data.
    normalize : bool
        if True the dataset will be normalized before training.
    test_size : float
        Between [0.0-1.0]. The size of the split for test within the training set.
    folds : int
        number of folds for cross validation.
    metric : str
        the metric used for evaluating the best model.
    model_name : str
        the name to save the model.
    testing_set : pd.DataFrame
        the external dataset for evaluating the best model.
    imbalanced
    seed : int
        random number to initilize the process.
    include_models : List
        a list of models to be included in the process.
    normalize_method : str
        The method used for normalizing the data.

    Returns
    -------
    Final regression model

    """
    if not metric:
        metric = 'RMSE'
    setup = pyreg.setup(target=y_col, data=training_set, normalize=normalize, normalize_method=normalize_method,
                        test_data=testing_set, fold=folds, session_id=seed,
                        feature_selection=feature_selection)
    best_model = pyreg.compare_models(sort=metric, include=include_models)
    pyreg.pull().to_csv(model_name + '_compare_models.tsv', sep='\t', index=False)
    reg_model = pyreg.create_model(best_model)
    reg_tuned_model = pyreg.tune_model(reg_model, optimize=metric)
    pyreg.pull().to_csv(model_name + '_tuned_model.tsv', sep='\t', index=False)
    final_model = pyreg.finalize_model(reg_tuned_model)
    pyreg.plot_model(final_model, save=True)
    pyreg.plot_model(final_model, plot='feature', save=True)
    pyreg.plot_model(final_model, plot='error', save=True)
    pyreg.save_model(final_model, model_name)
    if len(testing_set.index) != 0:
        unseen_predictions = test_regressor(
            model_path=model_name+'.pkl', x_set=testing_set.drop(columns=[y_col]), y_col=testing_set[y_col], output=model_name
        )
        unseen_predictions.to_csv(model_name + '_external_testing_results.tsv', sep='\t', index=True)
    return final_model


def classification_model(
    *,
    y_col,
    training_set,
    normalize,
    folds,
    metric,
    model_name,
    testing_set,
    imbalanced,
    seed,
    include_models,
    normalize_method,
    feature_selection
):
    """
    Build a classification model for prediction.

    Parameters
    ----------
    y_col : str
        the name of the target column.
    training_set : pd.DataFrame
        DataFrame containing the training data.
    normalize : bool
        if True the dataset will be normalized before training.
    test_size : float
        Between [0.0-1.0]. The size of the split for test within the training set.
    folds : int
        number of folds for cross validation.
    metric : str
        the metric used for evaluating the best model.
    model_name : str
        the name to save the model.
    testing_set : pd.DataFrame
        the external dataset for evaluating the best model.
    imbalanced : bool
        if True the imbalance will be fixed before the training.
    seed : int
        random number to initilize the process.
    include_models : List
        a list of models to be included in the process.
    normalize_method : str
        The method used for normalizing the data.

    Returns
    -------
    Final classification model

    """
    if not metric:
        metric = 'AUC'
    if len(training_set.groupby(y_col).groups) == 2:
        training_set[y_col] = np.interp(
            training_set[y_col], (training_set[y_col].min(), training_set[y_col].max()), (0, 1))
        testing_set[y_col] = np.interp(
            testing_set[y_col], (testing_set[y_col].min(), testing_set[y_col].max()), (0, 1))
    setup = pycl.setup(target=y_col, fix_imbalance=imbalanced, normalize=normalize, normalize_method=normalize_method,
                    data=training_set, test_data=testing_set, fold=folds, session_id=seed,
                    feature_selection=feature_selection)
    best_model = pycl.compare_models(sort=metric, include=include_models)
    pycl.pull().to_csv(model_name + '_compare_models.tsv', sep='\t', index=False)
    cl_model = pycl.create_model(best_model)
    cl_tuned_model = pycl.tune_model(cl_model, optimize=metric)
    pycl.pull().to_csv(model_name + '_tuned_model.tsv', sep='\t', index=False)
    final_model = pycl.finalize_model(cl_tuned_model)
    pycl.plot_model(final_model, plot='pr', save=True)
    pycl.plot_model(final_model, plot='feature', save=True)
    pycl.save_model(final_model, model_name)
    unseen_predictions = test_classifier(
        model_path=model_name+'.pkl', x_set=testing_set.drop(columns=[y_col]), y_col=testing_set[y_col], output=model_name
    )
    unseen_predictions.to_csv(model_name + '_external_testing_results.tsv', sep='\t', index=True)
    return final_model


def test_classifier(
    *,
    model_path,
    x_set,
    y_col,
    output
):
    """
    Test a model with an independent dataset.

    Parameters
    ----------
    model_path
        the model used for prediction
    x_set : pd.DataFrame
        the independent testing set (without the target)
    y_col : pd.Series
        the target true values.
    output : str
        the name for the evaluation output

    Returns
    -------
    pd.DataFrame
        the testing dataset with the true and predicted values.
    """
    model = joblib.load(model_path)
    x_set['Label'] = model.predict(x_set)
    x_set['True value'] = y_col
    if len(x_set.groupby('True value').groups) == 2:
        x_set['True value'] = np.interp(
            x_set['True value'], (x_set['True value'].min(), x_set['True value'].max()), (0, 1))
    report = metrics.classification_report(x_set['True value'], x_set['Label'])
    acc = metrics.accuracy_score(x_set['True value'], x_set['Label'])
    auc = metrics.roc_auc_score(x_set['True value'], x_set['Label'])
    input_list = [
        output, '\nTesting model report: \n', report + '\n', 'AUC = ' + str(auc) + '\n', 'Accuracy = ' + str(acc) + '\n'
    ]
    write_output(input_list=input_list, output=output + "_report.txt")
    generate_confusion_matrix(y_true=x_set['True value'], y_pred=x_set['Label'], output=output)
    return x_set


def test_regressor(
    *,
    model_path,
    x_set,
    y_col,
    output
):
    """
    Test a model with an independent dataset.

    Parameters
    ----------
    model_path
        the model used for prediction
    x_set : pd.DataFrame
        the independent testing set (without the target)
    y_col : pd.Series
        the target true values.
    output : str
        the name for the evaluation output

    Returns
    -------
    pd.DataFrame
        the testing dataset with the true and predicted values.

    """
    model = joblib.load(model_path)
    x_set['Label'] = model.predict(x_set)
    x_set['True value'] = y_col
    r2 = metrics.r2_score(y_col, x_set['Label'])
    rmse = metrics.mean_squared_error(y_col, x_set['Label'], squared=False)
    plot = generate_scatterplot(
        x_axis=x_set['Label'], y_axis=y_col, output=output)
    input_list = [output + '\nTesting model report: \n', 'R^2 = ' + str(r2) + '\n', 'RMSE = ' + str(rmse) + '\n']
    write_output(input_list=input_list, output=output + "_report.txt")
    return x_set



# -*- coding: utf-8 -*-
import gzip
import os
import re
import subprocess

import numpy as np
import pandas as pd
from scipy.stats import beta
from tqdm import tqdm

from .helpers import uni_profiles


def get_gene_info(
    *,
    annotation_file,
    variant_col,
    af_col,
    alt_col='Alt',
    del_col,
    output_dir,
    genes_col,
    maf_threshold=0.01,
    beta_param,
    weight_func='beta'
):
    """
    Create temporary files with variant information for each gene, plus the weights calculated.

    Parameters
    ----------
    annotation_file : str
        a file containing the variant, AF, ALT, Gene, and deleterious score.
    variant_col : str
        the name of the variant column.
    af_col : str
        the name of the Allele Frequency column.
    alt_col : str
        the name of the alternate allele column.
    del_col : str
        the name of functional annotation column.
    output_dir : str
        directory to save in temporary files.
    genes_col : str
        the name of genes column.
    maf_threshold : float
        between [0.0-1.0]. the minor allele frequency threshold, default is 0.01.
    beta_param : tuple
        the parameters of the beta function, if chosen for weighting.
    weight_func : str
        the weighting function, beta or log10.

    Returns
    -------
        output directory with all the temporary files.

    """
    # skip = 0
    # if annotated_vcf.endswith('.gz'):
    #     with gzip.open(annotated_vcf, 'r') as fin:
    #         for line in fin:
    #             if line.decode('utf-8').startswith('##'):
    #                 skip += 1
    # else:
    #     with open(annotated_vcf, 'r') as file:
    #         for line in file:
    #             if line.startswith('##'):
    #                 skip += 1
    df = pd.read_csv(annotation_file, sep=r'\s+', index_col=False, on_bad_lines='warn')
    # info = df['INFO'].str.split(pat=';', expand=True)
    # missing_info = info[info.isnull().all(axis=1)].index
    # df.drop(missing_info, inplace=True)
    # df.reset_index(drop=True, inplace=True)
    # info.drop(missing_info, inplace=True)
    # info.reset_index(drop=True, inplace=True)
    # for col in info.columns:
    #     val = str(info[col][0]).split('=')
    #     if len(val) == 1:
    #         continue
    #     info[col] = info[col].str.split(pat='=', expand=True).drop(columns=0)
    #     info.rename(columns={col: val[0]}, inplace=True)
    # df = pd.concat([df, info], axis=1)
    df[af_col].replace('.', 3.98e-6, inplace=True) # 1 allele out 125,748 indiv in gnomADexome (251496 alleles)
    df[af_col].replace('nan', 3.98e-6, inplace=True) # 1 allele out 125,748 indiv in gnomADexome (251496 alleles)
    df[del_col].replace('.', 0, inplace=True)
    df[del_col].replace('nan', 0, inplace=True)
    df = df[df[af_col].values.astype(float) <= maf_threshold]
    if weight_func == 'beta':
        df[weight_func] = beta.pdf(df[af_col].values.astype(float), beta_param[0], beta_param[1])
    elif weight_func == 'log10':
        df[weight_func] = -np.log10(df[af_col].values.astype(float))
        df[weight_func].replace([np.inf, -np.inf, np.nan], 0.0, inplace=True)
    df['score'] = df[weight_func].values.astype(float) * df[del_col].values.astype(float)
    genes = list(set(df[genes_col]))
    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    gene_file = output_dir + '.genes'
    with open(gene_file, 'w') as f:
        f.writelines("%s\n" % gene for gene in genes)
    [df[df[genes_col] == gene][[variant_col, alt_col, 'score', genes_col]].to_csv(os.path.join(output_dir, (
        str(gene) + '.w')), index=False, sep='\t') for gene in tqdm(genes, desc="writing w gene files")]
    [df[df[genes_col] == gene][[variant_col, alt_col]].to_csv(os.path.join(output_dir, (str(gene) + '.v')),
                                                              index=False, sep='\t') for gene in
     tqdm(genes, desc="writing v gene files")]
    return output_dir


def combine_scores(
    *,
    input_path,
    output_path,
    plink_extension,
):
    """
    Combine the files that contain the scores into one file.

    Parameters
    ----------
    input_path : str
        the directory containing scores files.
    output_path : str
        the name of the output file.

    Returns
    -------
    pd.DataFrame
        dataframe with all the scores.

    """
    all_files = [os.path.join(path, name) for path, subdirs, files in os.walk(input_path) for name in files]
    score_files = [f for f in all_files if re.match(r'.+'+plink_extension+'$', f)]
    try:
        df = pd.read_csv(str(score_files[0]), sep=r'\s+', on_bad_lines='warn').iloc[:, [1, -1]]
    except:
        raise Exception("It seems that there is a problem with your score files, please make sure that plink is running correctly.")
    scores_col = df.columns[1]
    df.astype({scores_col: np.float32})
    r = re.compile("([a-zA-Z0-9_.-]*)."+plink_extension+"$")
    gene = r.findall(str(score_files[0]))
    df.rename(columns={scores_col: gene[0]}, inplace=True)
    for i in tqdm(range(1, len(score_files)), desc='merging in process'):
        df = uni_profiles(df, score_files[i], plink_extension=plink_extension)
    df.to_csv(output_path, sep='\t', index=False)
    return df


def plink_process(*, genes_folder, plink, bfiles=None, method=""):
    """
    Use plink to calculate and sum the scores for each gene.

    Parameters
    ----------
    genes_folder : str
        the folder containing the temporary genes files.
    plink : str
        the directory of plink (if not default).
    bfiles : str
        the name of the binary files, not needed if annotated file is given.
    method : str
        if empty method is the average. if 'sum', then the variants scores will be summed.

    Returns
    -------

    """
    genes = [line.strip() for line in open(genes_folder + '.genes', 'r')]
    for gene in tqdm(genes, desc='calculating genes scores'):
        v_file = os.path.join(genes_folder, (gene + '.v'))
        w_file = os.path.join(genes_folder, (gene + '.w'))
        p = subprocess.call(
            plink + " --bfile " + bfiles + " --double-id" + " --extract " + v_file + " --score " + w_file +
            " 1 2 3 " + method + "--out " + os.path.join(genes_folder, gene), shell=True
        )


def calculate_gbrs(
    *,
    scores_df,
    weights_df,
    weights_col,
    genes_col,
    pathway_file=None,
    samples_col,
    method,
    logger
):
    """
    Calculate a gene-based risk score for each individual in a given dataset.

    Parameters
    ----------
    scores_df : pd.DataFrame
        the matrix with gene-based scores.
    weights_df : pd.DataFrame
        the matrix with weights for each gene.
    weights_col : str
        the name of the column with the weights.
    genes_col : str
        the name of the column with the genes.
    pathway_file : str
        if pathway method chosen, the file contains a list of pathways
    samples_col : str
        the name of the column with samples IDs.
    method : str
        the method used to calulate the gene-based risk scores. [sum | pathways]
    logger
        the logger to log the process

    Returns
    -------
    pd.DataFrame
        a dataframe with the gene-based risk scores.

    """
    genes = set(weights_df[genes_col]).intersection(set(scores_df.columns))
    df = scores_df[genes]
    df = df.reindex(sorted(df.columns), axis=1)
    df *= list(weights_df.sort_values(by=genes_col)[weights_col].values)
    if method == 'sum':
        df['gbrs'] = df.sum(axis=1)
        df = df[['gbrs']]
    elif method == 'pathways':
        pathways = {line.strip().split('\t')[0]: line.strip().split('\t')[2:] for line in open(pathway_file, 'r')}
        pathway_scores = pd.DataFrame(columns=[samples_col] + list(pathways))
        logger.info('calculating pathway scores ...')
        for path, path_genes in tqdm(pathways.items(), desc='calculating pathway scores'):
            selected_genes = list(set(genes) & (set(path_genes)))
            if len(selected_genes) == 0:
                pathway_scores.drop(columns=[path], inplace=True)
            pathway_scores[path] = df[selected_genes].sum(axis=1)/len(selected_genes)
        df = pathway_scores
    return df


def pathway_scoring(
    *,
    pathways,
    genes,
    scores_file,
    samples_col,
    logger
):
    """
    sum the gene-based scores to pathway-based scores.

    Parameters
    ----------
    pathways : dict
        the pathways (keys) and their genes (values)
    genes : List
        the genes to select from the gene-based scores.
    scores_file : str
        the gene-based scores file.
    samples_col : str
        the name the column of the samples IDs.
    logger
        the logger to log the process

    Returns
    -------
    pd.DataFrame
        dataframe containing the pathway-based scores.

    """
    logger.info('reading scores file ...')
    scores_df = pd.read_csv(scores_file, sep=r'\s+', usecols=[samples_col] + genes, on_bad_lines='warn')
    pathway_scores = pd.DataFrame(columns=[samples_col] + list(pathways))
    pathway_scores[samples_col] = scores_df[samples_col]
    logger.info('calculating pathway scores ...')
    for path, path_genes in tqdm(pathways.items(), desc='calculating pathway scores'):
        selected_genes = list(set(genes) & (set(path_genes)))
        if len(selected_genes) == 0:
            pathway_scores.drop(columns=[path], inplace=True)
        pathway_scores[path] = scores_df[selected_genes].sum(axis=1)/len(selected_genes)
    return pathway_scores



# -*- coding: utf-8 -*-
import os
import subprocess
from datetime import datetime

import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests

from .association_analysis import run_mannwhitneyu, run_ttest, get_pvals_logit, get_pvals_linear
from .gene_scoring import get_gene_info, plink_process, combine_scores
from .prediction_models import regression_model, classification_model, test_classifier, test_regressor
from .utils import gene_length_normalize

PATH = os.path.abspath(os.path.join((__file__), os.pardir, os.pardir, os.pardir))
BETAREG_SHELL = os.path.join(PATH, 'scripts', 'betareg_shell.R')
association_functions = {
    'mannwhitneyu': run_mannwhitneyu,
    'ttest_ind': run_ttest,
    'logit': get_pvals_logit,
    'linear': get_pvals_linear,
}


def scoring_process(
    *,
    logger,
    annotation_file,
    temp_dir,
    beta_param,
    weight_func,
    del_col,
    maf_threshold,
    gene_col,
    variant_col,
    af_col,
    alt_col,
    bfiles,
    plink,
    output_file,
):
    """
    Calculate gene-based scores.
    This is calculated by a weighted sum of the variants in a gene.

    Parameters
    ----------
    logger
        an object that logs function outputs.
    annotation_file : str
        an annotated containing variant IDs, alt, info and samples genotypes.
    temp_dir : str
        a temporary directory to save temporary files before merging.
    beta_param : tuple
        the parameters from beta weight function. ignore if log10 function is chosen.
    weight_func : str
        the weighting function used in score calculation.
    del_col : str
        the column containing deleteriousness score or functional annotation.
    maf_threshold : float
        the threshold for minor allele frequency. between [0.0-1.0]
    gene_col : str
        the column containing gene names.
    variant_col : str
        the column containing variant IDs.
    af_col : str
        the column containing allele frequency.
    alt_col : str
        the column containing alternate allele.
    bfiles : str
        the binary files for plink process. if vcf contains all information bfiles are not needed.
    plink : str
        the directory of plink, if not set in environment
    output_file : str
        the path to save the final output scores matrix.

    Returns
    -------
    DataFrame
        final scores matrix

    """
    try:
        genes_folder = get_gene_info(annotation_file=annotation_file, variant_col=variant_col, af_col=af_col, alt_col=alt_col,
                                     del_col=del_col, output_dir=temp_dir, genes_col=gene_col,
                                     maf_threshold=maf_threshold, beta_param=beta_param, weight_func=weight_func)
    except Exception as arg:
        logger.exception(arg)
        raise
    logger.info('calculating gene scores ...')
    try:
        plink_process(genes_folder=genes_folder, plink=plink, bfiles=bfiles)
    except Exception as arg:
        logger.exception(arg)
        raise
    logger.info('combining score files ...')
    try:
        if 'plink2' in plink:
            plink_extension = 'sscore'
        else:
            plink_extension = 'profile'
        df = combine_scores(input_path=temp_dir, output_path=output_file, plink_extension=plink_extension)
    except Exception as arg:
        logger.exception(arg)
        raise
    logger.info('process is complete.')
    return df


def find_pvalue(
    *,
    scores_file,
    info_file,
    genes=None,
    phenotype,
    samples_column,
    test='mannwhitneyu',
    covariates=None,
    cases=None,
    controls=None,
    processes=1,
    logger,
    zero_threshold=1.0,
    adj_pval,
):
    """
    Calculate the significance of a gene in a population using different statistical analyses [mannwhitneyu, logit, linear, ttest_ind]

    Parameters
    ----------
    scores_file : str
        dataframe containing the scores of genes across samples.
    info_file : str
        a file containing the information of the sample. this includes target phenotype and covariates.
    genes : list
        a list of the genes to calculate the significance. if None will calculate for all genes.
    phenotype : str
        the name of the column containing phenotype information.
    samples_column : str
        the name of the column contining samples IDs.
    test : str
        the type of statistical test to use, choices are: ttest_ind, mannwhitenyu, linear, logit.
    covariates : str
        the list of covariates used in the calculation.
    cases : str
        the cases category. if binary phenotype.
    controls : str
        the controls category. if binary phenotype.
    processes : int
        number of processes used, for parallel computing.
    logger
        an object that logs function outputs.

    Returns
    -------
    DataFrame
        dataframe with genes and their p_values

    """
    logger.info("Reading scores file...")
    if genes:
        scores_df = pd.read_csv(scores_file, sep=r'\s+',
                                usecols=lambda x: x in genes+[samples_column], on_bad_lines='warn')
    else:
        scores_df = pd.read_csv(scores_file, sep=r'\s+', on_bad_lines='warn')
    scores_df.set_index(samples_column, inplace=True)
    scores_df.replace([np.inf, -np.inf], 0, inplace=True)
    scores_df.fillna(0, inplace=True)
    scores_df = scores_df.loc[:, scores_df.var() != 0.0].reset_index()
    logger.info("Reading info file...")
    phenotypes_col = phenotype.split(',')
    if covariates:
        covariates = covariates.split(',')
        genotype_df = pd.read_csv(info_file, sep='\t', usecols=covariates+phenotypes_col+[samples_column], on_bad_lines='warn')
    else:
        genotype_df = pd.read_csv(info_file, sep='\t', on_bad_lines='warn')
    logger.info("Processing files...")
    merged_df = pd.merge(scores_df, genotype_df, how='inner', on=samples_column)
    merged_df.replace([np.inf, -np.inf], 0.0, inplace=True)
    genes = scores_df.columns.tolist()[1:]
    del scores_df
    args = {
        'processes': processes, 'cases': cases, 'controls': controls, 'covariates': covariates, 'logger': logger,
    }
    for pheno in phenotypes_col:
        logger.info("Calculating p_values using the following test: " + test + ' for the phenotype: ' + pheno)
        pheno_df = merged_df.dropna(subset=[pheno])
        sample_size = pheno_df.shape[0]
        if zero_threshold != 1.0:
            for gene in genes:
                df = pd.DataFrame(pheno_df[gene].value_counts()).reset_index()
                try:
                    zeros = df.loc[df['index'] == 0.0].iloc[0, 1]
                    freq = zeros / sample_size
                except:
                    freq = 0
                if freq >= zero_threshold:
                    pheno_df.drop(columns=gene, inplace=True)
                    genes.remove(gene)
        try:
            p_values_df = association_functions.get(test)(df=pheno_df, genes=genes, cases_column=pheno, **args)
            p_values_df.dropna(subset=['p_value'], inplace=True)
            p_values_df['p_value'].replace(0.0, 7.1429666685167604e-293, inplace=True)
            p_values_df['p_value'].replace(0, 7.1429666685167604e-293, inplace=True)
            if adj_pval:
                logger.info("Calculating the adjusted p_values...")
                adjusted = multipletests(list(p_values_df['p_value']), method=adj_pval)
                p_values_df[adj_pval + '_adj_pval'] = list(adjusted)[1]
            now = datetime.now().strftime("%Y%m%d")
            cohort = scores_file.split(r"/")[-1].split(".")[0]
            if zero_threshold != 1.0:
                output = now + "_" + test + "_" + str(int(zero_threshold * 100)) + "_" + cohort + "_" + pheno + ".tsv"
            else:
                output = now + "_" + test + "_" + cohort + "_" + pheno + ".tsv"
            p_values_df.to_csv(output, sep='\t', index=False)
        except Exception as arg:
            logger.exception(arg)
            raise
    return


def betareg_pvalues(
    *,
    scores_file,
    pheno_file,
    samples_col,
    cases_col,
    output_path,
    covariates,
    processes,
    genes,
    logger,
):
    """
    Calculate association significance using betareg. This function runs in Rscript.


    Parameters
    ----------
    scores_file : str
        the path to the scores file.
    pheno_file : str
        the path to the phenotypes and covariates file.
    samples_col : str
        column containing samples ids.
    cases_col : str
        column containing the phenotype.
    output_path : str
        a path to save the summary statistics.
    covariates : str
        the list of covariates used in the calculation.
    processes : int
        number of processes used, for parallel computing.
    genes : str
        a list of the genes to calculate the significance. if None will calculate for all genes.
    logger
        an object that logs function outputs.

    Returns
    -------

    """
    logger.info("The p_values will be calculated using beta regression.")
    if not genes:
        genes = ""
    try:
        p = subprocess.call(
            ["Rscript", BETAREG_SHELL,
             "-s", scores_file,
             "--phenofile", pheno_file,
             "--samplescol", samples_col,
             "--casescol", cases_col,
             "-o", output_path,
             "--covariates", covariates,
             "--processes", str(processes),
             "--genes", genes]
        )
    except Exception as arg:
        logger.exception(arg)
        raise
    logger.info("Process is complete. The association analysis results have been saved.")


def create_prediction_model(
    *,
    model_name='final_model',
    model_type='regressor',
    y_col,
    imbalanced=True,
    normalize=True,
    folds=10,
    training_set,
    testing_set=pd.DataFrame(),
    metric=None,
    seed,
    include_models,
    normalize_method,
    feature_selection
):
    """
    Create a prediction model (classifier or regressor) using the provided dataset.

    Parameters
    ----------
    model_name : str
        the name of the prediction model.
    model_type : str
        type of model [regressor|classifier]
    y_col : str
        the column containing the target (qualitative or quantitative).
    imbalanced : bool
        True means data is imbalanced.
    normalize : bool
        True if data needs normalization.
    folds : int
        how many folds for cross-validation.
    training_set : pd.DataFrame
        the training set for the model.
    testing_set : pd.DataFrame
        if exists an extra evaluation step will be done using the testing set.
    test_size : float
         the size to split the training set for cross-validation.
    metric : str
        the metric to evaluate the best model.
    seed : int
        the intilization state random number
    include_models : list
        a list of models that the user wants to test. if None all models will be used.
    normalize_method : str
        the method to normalize the data. Choices [zscore, minmax, maxabs, robust]

    Returns
    -------
    Final model

    """
    try:
        model_func = {'classifier': classification_model, 'regressor': regression_model}
        final_model = model_func.get(model_type)(
            y_col=y_col,
            training_set=training_set,
            normalize=normalize,
            folds=folds,
            metric=metric,
            model_name=model_name,
            testing_set=testing_set,
            imbalanced=imbalanced,
            seed=seed,
            include_models=include_models,
            normalize_method=normalize_method,
            feature_selection=feature_selection
        )
    except Exception:
        raise Exception('Model requested is not available. Please choose regressor or classifier.')
    return final_model


def model_testing(
    *,
    model_path,
    input_file,
    samples_col,
    label_col,
    model_type,
    output
):
    """
    Load a prediction model and use it to predict label values in an independent dataset.

    Parameters
    ----------
    model_path : str
        the path to saved model.
    input_file : str
        the file with the test dataset.
    samples_col : str
        the name of samples column.
    label_col : str
        the name of the target column
    model_type :
        regressor or classifier

    Returns
    -------
    DataFrame
        the results of predictions.

    """
    testing_df = pd.read_csv(input_file, sep='\t', index_col=samples_col, on_bad_lines='warn')
    testing_df = testing_df.dropna(subset=[label_col])
    testing_df.replace([np.inf, -np.inf, np.nan], 0.0, inplace=True)
    x_set = testing_df.drop(columns=label_col)
    model_func = {'classifier': test_classifier, 'regressor': test_regressor}
    unseen_predictions = model_func.get(model_type)(
        y_col=testing_df[label_col],
        output=output,
        model_path=model_path,
        x_set=x_set
    )
    return unseen_predictions


def normalize_data(
    *,
    method='gene_length',
    genes_info=None,
    genes_col='HGNC symbol',
    length_col='gene_length',
    data_file,
    samples_col,
):
    """
    Normalize dataset using gene_length, minmax, maxabs, zscore or robust

    Parameters
    ----------
    method : str
        the normalization method. [zscore, gene_length, minmax, maxabs, robust]
    genes_info : str
        file containing the genes and their lengths. if gene_length method chosen with no file, info will be retrieved from ensembl database.
    genes_col : str
        the column containing genes (if genes_info file is provided)
    length_col : str
        the columns containing genes length (if genes_info file is provided)
    data_file : str
        file containing dataset to be normalized.
    samples_col : str
        the column containing samples ids.

    Returns
    -------
    DataFrame
        a df with the normalized dataset.

    """
    scores_df = pd.read_csv(data_file, sep=r'\s+', on_bad_lines='warn')
    if method == 'gene_length':
        scores_df = gene_length_normalize(
            genes_info=genes_info, genes_col=genes_col, length_col=length_col, scores_df=scores_df,
            samples_col=samples_col
        )
    elif method == 'maxabs':
        for col in scores_df.columns:
            if col == samples_col:
                continue
            scores_df[col] = scores_df[col]/scores_df[col].abs().max()
    elif method == 'minmax':
        for col in scores_df.columns:
            if col == samples_col:
                continue
            scores_df[col] = (scores_df[col] - scores_df[col].min()) / (scores_df[col].max() - scores_df[col].min())
    elif method == 'zscore':
        scores_df.std(ddof=0)
        for col in scores_df.columns:
            if col == samples_col:
                continue
            scores_df[col] = (scores_df[col] - scores_df[col].mean()) / scores_df[col].std()
    elif method == 'robust':
        for col in scores_df.columns:
            if col == samples_col:
                continue
            scores_df[col] = (scores_df[col] - scores_df[col].median()) / (scores_df[col].quantile(0.75) - scores_df[col].quantile(0.25))
    else:
        raise Exception(
            'This function does not support the normalization method you selected. Methods: [zscore, gene_length, minmax, maxabs, robust]'
        )
    scores_df.replace([np.inf, -np.inf], 0, inplace=True)
    return scores_df



# -*- coding: utf-8 -*-
from datetime import datetime
import re

import logging
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn.metrics as metrics


def uni_profiles(df, f, plink_extension):
    """
    Merge two dataframes.

    Parameters
    ----------
    df : pd.DataFrame
        the main dataframe with all the scores.
    f : str
        the file containing the scores of one gene.

    Returns
    -------
    pd.DataFrame
        the merged dataframe

    """
    df2 = pd.read_csv(str(f), sep=r'\s+').iloc[:, [1, -1]]
    scores_col = df2.columns[1]
    df2.astype({scores_col: np.float32})
    r = re.compile("([a-zA-Z0-9_.-]*)."+plink_extension+"$")
    gene2 = r.findall(str(f))
    df2.rename(columns={scores_col: gene2[0]}, inplace=True)
    df = pd.merge(df, df2, on='IID')
    return df


def generate_scatterplot(*, x_axis, y_axis, output):
    """
    Generate scatterplot.

    Parameters
    ----------
    x_axis : str
        the name of the x-axis column.
    y_axis : str
        the name of the y-axis columns.
    output : str
        the output image path.

    Returns
    -------
    a scatterplot

    """
    plt.scatter(x_axis, y_axis, alpha=0.5)
    m, b = np.polyfit(x_axis, y_axis, 1)
    plt.plot(x_axis, m * x_axis + b, 'r')
    plt.title('Actual vs predicted scatterplot')
    plt.xlabel('Predicted')
    plt.ylabel('Actual')
    plt.savefig(output + '_regressor_scatterplot.png')
    return plt.show()


def generate_confusion_matrix(y_true, y_pred, output):
    """
    Generate a confusion matrix for a dataset.

    Parameters
    ----------
    x_set : pd.DataFrame
        the features set.
    y_set : pd.Series
        the target set.
    output : str
        the output image path

    Returns
    -------
        confusion matric plot

    """
    cm = metrics.confusion_matrix(y_true, y_pred)
    cm_display = metrics.ConfusionMatrixDisplay(confusion_matrix=cm).plot()
    cm_display.ax_.set_title('Classifier confusion matrix')
    plt.savefig(output + '_classifier_confusion_matrix.png')
    return


def write_output(*, input_list, output):
    """
    Write list info a txt file.

    Parameters
    ----------
    input_list : List
        a list of lines
    output : str
        the output file

    Returns
    -------

    """
    textfile = open(output, "w")
    for line in input_list:
        textfile.write(line)
    textfile.close()


def create_logger():
    """
    Create a logger to output command information.

    Returns
    -------
        logger

    """
    logger = logging.getLogger('genrisk')
    logger.setLevel(logging.DEBUG)
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    now = datetime.now()
    log_file = 'genrisk_' + now.strftime("%d%m%Y_%H%M") + '.log'
    # create file handler which logs even debug messages
    fh = logging.FileHandler(log_file)
    fh.setLevel(logging.DEBUG)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)
    return logger



# -*- coding: utf-8 -*-
import os
import random
import shutil
import time

import click
import pandas as pd
import numpy as np

from sklearn.model_selection import train_test_split
from statsmodels.stats.multitest import multipletests

from .gene_scoring import calculate_gbrs, pathway_scoring
from .helpers import create_logger
from .pipeline import find_pvalue, betareg_pvalues, create_prediction_model, model_testing, scoring_process, \
    normalize_data
from .prs_scoring import prs_prompt
from .utils import draw_qqplot, draw_manhattan, merge_files

SAMPLES_COL = click.option('-m', '--samples-col',  default='IID', show_default=True,
                           help="the name of the column that contains the samples.")
OUTPUT_FILE = click.option('-o', '--output-file', required=True, help="the final output path")
logger = create_logger()


@click.group()
def main():
    """Handle genrisk functions."""


@main.command()
@click.option('-a', '--annotation-file', required=True, type=click.Path(exists=True), help='an annotation file containing variant IDs, alt, AF and deletarious scores.')
@click.option('-b', '--bfiles', required=True,
              help='provide binary files that contain the samples info')
@click.option('--plink', default='plink', help="the directory of plink, if not set in environment")
@click.option('-t', '--temp-dir', required=True, help="a temporary directory to save temporary files before merging.")
@OUTPUT_FILE
@click.option('-p', '--beta-param', show_default=True, default=(1.0, 25.0), nargs=2, type=float,
              help="the parameters from beta weight function.")
@click.option('-w', '--weight-func', show_default=True, default='beta', type=click.Choice(['beta', 'log10']),
              help="the weighting function used in score calculation.")
@click.option('-v', '--variant-col', show_default=True, default='SNP', help="the column containing the variant IDs.")
@click.option('-g', '--gene-col', show_default=True, default='Gene.refGene', help="the column containing gene names.")
@click.option('-f', '--af-col', show_default=True, default='MAF', help="the column containing allele frequency.")
@click.option('-d', '--del-col', show_default=True, default='CADD_raw', help="the column containing the deleteriousness score.")
@click.option('-l', '--alt-col', show_default=True, default='Alt', help="the column containing the alternate base.")
@click.option('-m', '--maf-threshold', show_default=True, default=0.01, help="the threshold for minor allele frequency.")
@click.option('-k', '--keep', is_flag=True, help='if flagged temporary files will not be deleted.')
def score_genes(
        *,
        annotation_file,
        bfiles,
        plink,
        beta_param,
        temp_dir,
        output_file,
        weight_func,
        variant_col,
        gene_col,
        af_col,
        del_col,
        alt_col,
        maf_threshold,
        keep
):
    """
    Calculate the gene-based scores for a given dataset.

    Example
    ---------
    ::

        $ genrisk score-genes -a /path/to/toy_vcf_info.vcf -o toy_genes_scores.tsv -t toy_vcf_scoring -v ID -f AF -g gene -l ALT -d RawScore

    \f

    Parameters
    ----------
    annotation_file : str
        an annotation file containing variant IDs, alt, AF and deleterious scores.
    bfiles : str
        the binary files for plink process.
    plink : str
        the location of plink, if not set in environment
    beta_param : tuple
        the parameters from beta weight function.
    temp_dir : str
        a temporary directory to save temporary files before merging.
    output_file : str
        the location and name of the final output scores matrix.
    weight_func : str
        the weighting function used on allele frequency in score calculation. [beta| log10]
    variant_col : str
        the column containing the variant IDs.
    gene_col : str
        the column containing gene names.
        If the genes are in the INFO column, use the identifier of the value (i.e gene=IF, identifier is 'gene')
    af_col : str
        the column containing allele frequency. If in INFO, follow previous example
    del_col : str
        the column containing deleteriousness score (functional annotation). If in INFO, follow previous example
    alt_col : str
        the column containing alternate base.
    maf_threshold : float
        the threshold for minor allele frequency.

    Returns
    -------
    DataFrame information
        the final scores dataframe information
        the DataFrame is saved into the output path indicated in the arguments

    """
    logger.info('Gene-based scoring')
    logger.info('Score genes process is starting now...')
    logger.info(locals())
    logger.info('getting information from vcf files')
    start_time = time.time()
    df = scoring_process(logger=logger, annotation_file=annotation_file, temp_dir=temp_dir, beta_param=beta_param,
                         weight_func=weight_func, del_col=del_col, maf_threshold=maf_threshold, gene_col=gene_col,
                         variant_col=variant_col, af_col=af_col, alt_col=alt_col, bfiles=bfiles, plink=plink,
                         output_file=output_file)
    if not keep:
        logger.info('The temporary files will be removed now.')
        shutil.rmtree(temp_dir)
    time.sleep(1)
    end_time = time.time()
    logger.info(f"Runtime of the program is {end_time - start_time} Sec")
    return df.info()


@main.command()
@click.option('-s', '--scores-file', required=True, type=click.Path(exists=True),
              help="The scoring file of genes across a population.")
@click.option('-i', '--info-file', required=True, type=click.Path(exists=True),
              help="File containing information about the cohort.")
@click.option('-g', '--genes', default="",
              help="a file containing the genes to calculate. if not provided all genes will be used.")
@click.option('-t', '--test', required=True,
              type=click.Choice(['ttest_ind', 'mannwhitneyu', 'logit', 'linear']),
              help='statistical test for calculating P value.')
@click.option('-c', '--phenotype', required=True,
              help="the name of the column that contains the case/control or quantitative vals.")
@SAMPLES_COL
@click.option('-a', '--adj-pval', type=click.Choice(
    ['bonferroni', 'sidak', 'holm-sidak', 'holm',
     'simes-hochberg', 'hommel', 'fdr_bh', 'fdr_by', 'fdr_tsbh', 'fdr_tsbky']))
@click.option('-v', '--covariates', default='', help="the covariates used for calculation")
@click.option('-p', '--processes', show_default=True, type=int, default=1, help='number of processes for parallelization')
@click.option('--zero-threshold', type=float, default=1.0, help='the threshold for the frequency of zeros per gene to be included')
def find_association(
        *,
        scores_file,
        info_file,
        genes,
        phenotype,
        samples_col,
        test,
        adj_pval,
        covariates,
        processes,
        zero_threshold,
):
    """
    Calculate the P-value between two given groups.

    Example
    ---------
    ::

        $ genrisk find-association --scores-file toy_example/toy_dataset_scores --info-file
        toy_example/toy.pheno --phenotype trait1,trait2 --samples-column IID --test logit
         --covariates age,sex --adj-pval bonferroni
    \f

    Parameters
    ----------
    scores_file : str
        the file containing gene-based scores.
    info_file : str
        file containing the phenotype.
    genes : str
        a file that contains a list of genes to calculate p-values. if not, all genes in scoring file will be used.
    phenotype : str
        the name of the column with phenotypes. Phenotypes can be either binary or quantitative.
    samples_col : str
         the name of the column with sample IDs. All files need to have the same format.
    test : str
        the statistical test used for calculating p-values.
    adj_pval : str, optional
        the method used to adjust the p-values.
    covariates : str, optional
        the covariates used for calculation. Not all tests are able to include covariates.
        (e.g. Mann Whinteny U doesn't allow for covariates)
    processes : int, optional
        if more than 1 processer is selected, the function will be parallelized.

    Returns
    -------
    DataFrame information
        the final dataframe information
        the DataFrame is saved into the output path indicated in the arguments

    """
    logger.info('Finding associations')
    logger.info(locals())
    logger.info("The process for calculating the p_values will start now.")
    start_time = time.time()
    # if test == 'betareg':
    #     betareg_pvalues(
    #         scores_file=scores_file,
    #         pheno_file=info_file,
    #         cases_col=phenotype,
    #         samples_col=samples_col,
    #         output_path=output_file,
    #         covariates=covariates,
    #         processes=processes,
    #         genes=genes,
    #         logger=logger,
    #     )
    #     end_time = time.time()
    #     logger.info(f"Runtime of the program is {end_time - start_time}")
    #     df = pd.read_csv(output_file, sep='\t', index_col=False)
    #     if adj_pval:
    #         logger.info("Calculating the adjusted p_values...")
    #         adjusted = multipletests(list(df['p_value']), method=adj_pval)
    #         df[adj_pval + '_adj_pval'] = list(adjusted)[1]
    #     df.to_csv(output_file, sep='\t', index=False)
    if genes:
        with open(genes) as f:
            content = f.readlines()
        genes = [x.strip() for x in content]
    df = find_pvalue(
        scores_file=scores_file,
        info_file=info_file,
        genes=genes,
        phenotype=phenotype,
        samples_column=samples_col,
        test=test,
        covariates=covariates,
        processes=processes,
        logger=logger,
        zero_threshold=zero_threshold,
        adj_pval=adj_pval,
    )
    end_time = time.time()
    logger.info(f"Runtime of the program is {end_time - start_time}")
    logger.info("Process is complete. The association analysis results have been saved.")


@main.command()
@click.option('-p', '--pvals-file', required=True, type=click.Path(exists=True), help="the file containing p-values.")
@click.option('-i', '--info-file', type=click.Path(exists=True), help="file containing variant/gene info.")
@click.option('--genescol-1', show_default=True, default='genes', help="the name of the genes column in pvals file.")
@click.option('--genescol-2', show_default=True, default='Gene.refGene', help="the name of the genes column in info file.")
@click.option('-v', '--pval-col', show_default=True, default='p_value', help="the name of the pvalues column.")
@click.option('-c', '--chr-col', show_default=True, default='Chr', help='the name of the chromosomes column')
@click.option('-s', '--pos-col', show_default=True, default='Start', help='the name of the position/start of the gene column')
def visualize(
        *,
        pvals_file,
        info_file,
        genescol_1,
        genescol_2,
        pval_col,
        chr_col,
        pos_col,
):
    """
    Visualize manhatten plot and qqplot for the data.

    Example
    --------
    ::

        $ genrisk visualize --pvals-file toy_example/toy_dataset_scores
        --info-file annotated_toy_dataset.vcf
    \f

    Parameters
    ----------
    pvals_file : str
        the file containing the calculated p-values.
    info_file : str
        file containing variant/gene info.
    genescol_1 : str
        the name of the genes column in pvals file.
    genescol_2 : str
        the name of the genes column in info file.
    pval_col : str
        the name of the pvalues column.
    chr_col : str
        the name of chromosomes column.
    pos_col : str
        the name of the position/start column.

    Returns
    -------

    """
    logger.info('Creating plots for data')
    logger.info(locals())
    logger.info('Reading p_values file...')
    pvals_df = pd.read_csv(pvals_file, sep='\t', index_col=False)
    prefix = pvals_file.split(".")[0]
    qq_output= prefix + "_qqplot.png"
    manhattan_output = prefix + "_manhattan.png"
    logger.info('Creating QQ-plot...')
    try:
        draw_qqplot(pvals=pvals_df[pval_col], qq_output=qq_output)
    except Exception as arg:
        logger.exception(arg)
        raise
    logger.info('Creating Manhattan plot...')
    if not info_file:
        logger.exception('Please provide a file with gene information to generate manhattan plot.')
    info_df = pd.read_csv(info_file, sep="\t", index_col=False)
    merged = pd.merge(pvals_df, info_df, left_on=genescol_1, right_on=genescol_2, how='left')
    try:
        draw_manhattan(
            data=merged,
            chr_col=chr_col,
            pos_col=pos_col,
            pvals_col=pval_col,
            genes_col=genescol_1,
            manhattan_output=manhattan_output
        )
    except Exception as arg:
        logger.exception(arg)
        raise
    logger.info('Done.')


@main.command()
@click.option('-d', '--data-file', type=click.Path(exists=True), required=True,
              help='file with all features and target for training model.')
@click.option('-o', '--output-folder', required=True, help='path of folder that will contain all outputs.')
@click.option('-i', '--test-size', show_default=True, default=0.25, help='test size for cross validation and evaluation.')
@click.option('-n', '--model-name', required=True, help='name of model file.')
@click.option('--model-type', required=True, type=click.Choice(['regressor', 'classifier']),
              help='type of prediction model.')
@click.option('-l', '--target-col', required=True, help='name of target column in data_file.')
@click.option('-b', '--imbalanced', is_flag=True, help='if flagged methods will be used to account for the imbalance.')
@click.option('--normalize', is_flag=True, help='if flagged the data will be normalized before training.')
@click.option('--normalize-method', show_default=True, default='zscore', type=click.Choice(['zscore', 'minmax', 'maxabs', 'robust']),
              help='features normalization method.')
@click.option('-f', '--folds', show_default=True, default=10, type=int, help='number of cross-validation folds in training.')
@click.option('--metric', help='the metric used to choose best model after training.')
@SAMPLES_COL
@click.option('--seed', default=random.randint(1, 2147483647),
              help='add number to create reproduciple train_test splitting.')
@click.option('--include-models', default=None,
              help='choose specific models to compare with comma in between. e.g lr,gbr,dt')
@click.option('--feature-selection', is_flag=True,
              help='if selected feature selection will be implemented in training.')
def create_model(
        *,
        data_file,
        output_folder,
        test_size=0.25,
        model_name,
        model_type,
        target_col,
        imbalanced,
        normalize,
        normalize_method,
        folds,
        metric,
        samples_col,
        seed,
        include_models,
        feature_selection,
):
    """
    Create a prediction model with given dataset.

    Example
    --------
     ::

        $ genrisk create-model --data-file toy_example_regressor_features.tsv --model-type regressor
        --output-folder toy_regressor  --test-size 0.25 --test --model-name toy_regressor
        --target-col trait1 --imbalanced --normalize

    Notes
    -------
    The types of models available for training can be found :ref:`model_types`
    \f

    Parameters
    ----------
    data_file : str
        file containing features and target.
    output_folder : str
        a folder path to save all outputs.
    test_size : float
        the size of testing set.
    test : bool
        if True the dataset will be split into training and testing for extra evaluation after finalization.
    model_name : str
        the name of the model to be saved.
    model_type : str
        the type of model [regressor| classifier].
    target_col : str
        the name of the target column in data file.
    imbalanced : bool
        if true methods will be used to account for the imbalance.
    normalize : bool
        if true the data will be normalized before training
    normalize_method : str
        method used to normalize data. [zscore| minmax| maxab| robust]
    folds : int
        the number of folds used for cross validation
    metric : str
        the metric used to choose best model after training.
    samples_col : str
        the name of the column with samples IDs.
    seed : int
        random seed number to run the machine learning models.
    include_models : str
        list of specific models to compare. more information in the documentations

    Returns
    -------
    Final prediction model

    """
    logger.info('Create prediction model')
    logger.info(locals())
    logger.info('Reading and preparing data ...')
    start_time = time.time()
    data = pd.read_csv(data_file, sep='\t', index_col=samples_col)
    data.dropna(subset=[target_col], inplace=True)
    data.replace([np.inf, -np.inf, np.nan], 0.0, inplace=True)
    training_set, testing_set = train_test_split(data, test_size=test_size, random_state=int(seed))
    os.mkdir(output_folder)
    os.chdir(output_folder)
    logger.info('Model generation starting ...')
    if include_models:
        include_models = include_models.split(',')
    model = create_prediction_model(
        model_name=model_name,
        model_type=model_type,
        imbalanced=imbalanced,
        normalize=normalize,
        folds=folds,
        metric=metric,
        y_col=target_col,
        training_set=training_set,
        testing_set=testing_set,
        seed=int(seed),
        include_models=include_models,
        normalize_method=normalize_method,
        feature_selection=feature_selection
    )
    logger.info('Model is generated.')
    end_time = time.time()
    logger.info(f"Runtime of the program is {end_time - start_time}")
    return model


@main.command()
@click.option('-t', '--model-type', required=True, type=click.Choice(['regressor', 'classifier']),
              help='type of prediction model.')
@click.option('-i', '--input-file', required=True, type=click.Path(exists=True), help='testing dataset')
@click.option('-l', '--label-col', required=True, help='the target/phenotype/label column')
@click.option('-m', '--model-path', required=True, type=click.Path(exists=True), help='path to the trained model.')
@click.option('-s', '--samples-col', show_default=True, default='IID', help='the samples column.')
@OUTPUT_FILE
def test_model(
        *,
        model_path,
        input_file,
        model_type,
        label_col,
        samples_col,
        output_file,
):
    """
    Evaluate a prediction model with a given dataset.

    Example
    --------
     ::

        $ genrisk test-model --model-path regressor_model.pkl --input-file testing_dataset.tsv
        --model-type regressor --labels-col target --samples-col IID
    \f

    Parameters
    ----------
    model_path : str
        the path to the ML model.
    input_file : str
        the testing (independent) dataset.
    model_type : str
        the type of model [classifier|regressor].
    label_col : str
        the labels/target column.
    samples_col : str
        the sample ids column.
    output_file : str
        the path to the dataframe with the prediction results.

    Returns
    -------
    DataFrame
        dataframe with the prediction results.

    """
    logger.info('Testing prediction model')
    logger.info(locals())
    testing_df = model_testing(
        model_path=model_path,
        input_file=input_file,
        label_col=label_col,
        samples_col=samples_col,
        model_type=model_type,
        output=output_file.split('.')[0],
    )
    logger.info('saving test predictions')
    testing_df.to_csv(output_file, sep='\t')
    return testing_df


@main.command()
@click.option('-p', '--plink', default='plink')
def get_prs(
        *,
        plink,
):
    """
    Calculate PRS. This command is interactive.
    This command gets a pgs file (provided by the user or downloaded) then calculates the PRS for dataset.

    Example
    -------
    This function is performed using commandline interface::

        $ genrisk get-prs
    \f

    Parameters
    ----------
    plink : str
        provide plink path if not default in environment.

    Returns
    -------

    """

    download = click.confirm('Do you want to download PGS file?')
    return prs_prompt(plink=plink, download=download)


@main.command()
@click.option('-f', '--files', required=True,
              help='input all files to merge with a comma in between. E.g: file1,file2,file3')
@click.option('-s', '--sep', show_default=True, default=r'\s+', help='the column seperator in files.')
@click.option('-b', '--by', show_default=True, default='IID', help='the common column between all files to merge.')
@click.option('-c', '--cols', default=None,
              help='if desired, a list of columns can be chosen to save final file, e.g: col1,col2,col5')
@OUTPUT_FILE
def merge(
        *,
        files,
        sep,
        by,
        cols,
        output_file
):
    """
    Merge all files given into one dataframe. \f

    Parameters
    ----------
    files : str
         all files to merge with a comma in between. E.g: file1,file2,file3
    sep : str
        the column seperator in files.
    by : str
        the common column between all files to merge.
    cols : str
        a list of columns can be chosen to save final file, e.g: col1,col2,col5. if empty all columns will be merged
    output_file : str
        the name and path to save final dataframe

    Returns
    -------
    DataFrame
        final merged dataframe.

    """
    logger.info('GenRisk - merging dataframes')
    logger.info(locals())
    files_lst = files.split(',')
    df = merge_files(
        files_lst=files_lst,
        sep=sep,
        by=by,
        cols=cols
    )
    logger.info('saving merged dataframe.')
    df.to_csv(output_file, sep='\t')
    return df


@main.command()
@click.option('-s', '--scores-file', required=True, type=click.Path(exists=True), help='the gene-based scores file.')
@click.option('-w', '--weights-file', default=None, type=click.Path(exists=True),
              help='a file containg the weights for each gene.')
@click.option('-p', '--pheno-file', default=None, type=click.Path(exists=True),
              help='if no weights are given, pheno file is used to calculate them.')
@click.option('-c', '--pheno-col', default=None, help='the column containing the phenotype.')
@click.option('-v', '--covariates', show_default=True, default='sex,age,bmi,PC1,PC2,PC3,PC4',
              help='the covariates to use in the linear model.')
@click.option('-g', '--genes-col', show_default=True, default='genes', help='the column containing gene names.')
@click.option('-e', '--weights-col', show_default=True, default='zscore', help='the name and path to output results.')
@OUTPUT_FILE
@click.option('--split-size', show_default=True, default=0.25, help='the size ratio to split dataset for weight calculation.')
@click.option('--method', type=click.Choice(['sum', 'pathways']),
              help='method for presenting the risk scores.')
@SAMPLES_COL
@click.option('--pathway-file', default=None, help='.gmt file containing the pathway and its genes.')
def get_gbrs(
        *,
        method,
        scores_file,
        weights_file,
        pheno_file,
        pheno_col,
        covariates,
        genes_col,
        samples_col,
        weights_col,
        output_file,
        split_size,
        pathway_file
):
    """
    Calculate gene-based risk score for individuals. 
    \f

    Parameters
    ----------
    method : str
        for presenting the risk score, either sum of all genes or pathways.
    scores_file : str
        the gene-based scores file
    weights_file : str
        a file containing the weight for each gene.
    pheno_file : str
        if no weights are given, pheno file is used to calculate them
    pheno_col : str
        the column containing the phenotype.
    covariates : str
        the covariates to use in the linear model.
    genes_col : str
        column containing the gene names.
    samples_col : str
        column containing the samples ids.
    weights_col : str
        column containing the weights for each gene (in the weights file)
    output_file : str
        the name and path to output results.
    split_size : str
        the proportion of the dataset to use for calculating the weights. should be between 0.0 and 1.0.
    pathway_file : str
        if pathways method is chosen, a file with the pathways is needed

    Returns
    -------
    DataFrame
        df contianing the final risk scores.

    """
    logger.info('GenRisk - Calculating gene-based risk scores')
    logger.info(locals())
    logger.info("Reading files...")
    scores_df = pd.read_csv(scores_file, sep=r'\s+', index_col=False)
    if weights_file:
        weights_df = pd.read_csv(weights_file, sep='\t')
    else:
        logger.info("No weights available, weights will be caluclated now.")
        logger.info("Creating temporary files...")
        scores_df, scores_temp = train_test_split(scores_df, test_size=split_size, random_state=1)
        scores_temp.to_csv('scores_temp.tsv', sep='\t', index=False)
        logger.info("The process for calculating the p_values will start now.")
        weights_df = find_pvalue(
            scores_file='scores_temp.tsv',
            info_file=pheno_file,
            phenotype=pheno_col,
            samples_column=samples_col,
            test='linear',
            covariates=covariates,
            logger=logger,
            zero_threshold=1.0,
        )
        weights_df.to_csv(output_file+'.assoc', sep='\t', index=False)
        logger.info("Exluding samples used in weights calculation.")
        scores_df.reset_index(drop=True, inplace=True)
        logger.info("Remove temporary files.")
        del scores_temp
        os.remove('scores_temp.tsv')
        weights_df['zscore'] = weights_df['beta_coef'] / weights_df['std_err']
        weights_df.to_csv(output_file+'.weights', sep='\t', index=False)
    logger.info("Calculating GBRS now ...")
    weights_df[weights_col] = pd.to_numeric(weights_df[weights_col], errors='coerce')
    df = calculate_gbrs(
        scores_df=scores_df,
        weights_df=weights_df,
        weights_col=weights_col,
        genes_col=genes_col,
        method=method,
        pathway_file=pathway_file,
        samples_col=samples_col,
        logger=logger,
    )
    df[samples_col] = scores_df[samples_col]
    logger.info("GBRS dataframe is being saved ...")
    df.to_csv(output_file, sep='\t', index=False)
    logger.info("Process is complete.")
    return df


@main.command()
@OUTPUT_FILE
@SAMPLES_COL
@click.option('-p', '--pathway-file', required=True, type=click.Path(exists=True),
              help='.gmt file containing the pathway and its genes.')
@click.option('-s', '--scores-file', required=True, type=click.Path(exists=True),
              help='genes scores file to calculate pathway scores.')
def calculate_pathways(
        *,
        output_file,
        pathway_file,
        scores_file,
        samples_col
):
    """
    Calculate pathway scores using gene-based scores and gmt pathway file. \f

    Parameters
    ----------
    output_file : str
        the path to final output.
    pathway_file : str
        .gmt file containing the pathway and its genes.
    scores_file : str
        genes scores file to calculate pathway scores.
    samples_col : str
        column containing samples ids.

    Returns
    -------
    DataFrame
        the pathway scores df

    """
    logger.info('GenRisk - calculating pathway scores')
    logger.info(locals())
    pathways = {line.strip().split('\t')[0]: line.strip().split('\t')[2:] for line in open(pathway_file, 'r')}
    all_genes = [item for sublist in list(pathways.values()) for item in sublist]
    scored_genes = open(scores_file).readline().rstrip().split()
    combined_genes = list(set(all_genes) & set(scored_genes))
    df = pathway_scoring(pathways=pathways, genes=combined_genes, scores_file=scores_file, samples_col=samples_col,
                         logger=logger)
    df.to_csv(output_file, sep='\t', index=False)
    logger.info('Process is done.')
    return df


@main.command()
@click.option('--method', required=True, type=click.Choice(['gene_length', 'zscore', 'minmax','maxabs', 'robust']))
@click.option('--data-file', required=True, type=click.Path(exists=True))
@click.option('--genes-info', default=None)
@SAMPLES_COL
@click.option('--genes-col', show_default=True, default='HGNC symbol', type=str)
@click.option('--lengths-col', show_default=True, default='gene_length', type=str)
@OUTPUT_FILE
def normalize(
        *,
        method,
        data_file,
        genes_info,
        samples_col,
        genes_col,
        lengths_col,
        output_file,

):
    """
    Normalize/standarize data.

    Example
    -------
    ::

        $ genrisk normalize --data-file toy_example/toy_dataset_scores --method gene_length --samples-col IID
        --output-file toy_dataset_scores_normalized.tsv

    \f

    Parameters
    ----------
    genes_info : str
        the file containing genes names and length. if not provided ensembl database is used to retrieve data.
    method : str
        the method of normalizing data. [gene_length|zscore|minmax|maxabs|robust]
    data_file : str
        the file containg data to be normalized.
    samples_col : str
        the column containing sample ids.
    genes_col : str
        the column containing gene names. ignore if genes_info file is not provided.
    lengths_col : str
        the column containing gene lengths. ignore if genes_info file is not provided.
    output_file : str
        the name of the file for final output

    Returns
    -------
    DataFrame with normalized data.

    """
    logger.info('GenRisk - normalizing data using '+ method)
    logger.info(locals())
    df = normalize_data(
        method=method, data_file=data_file, samples_col=samples_col, genes_col=genes_col, length_col=lengths_col,
        genes_info=genes_info,
    )
    df.to_csv(output_file, sep='\t', index=False)
    logger.info('Process is done.')
    return df

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from subprocess import call, run
from urllib import request

import click
from requests import get


def prs_prompt(*, download, plink):
    """
    Interactive function to download and calculate prs.

    Parameters
    ----------
    download : bool
        if true, PGS file will be downloaded given an ID
    plink : str
        the path to plink. if not default in environment.

    Returns
    -------
        a statement
    """
    if download:
        pgs_id = click.prompt('Please input PGS ID', type=str)
        pgs_file = download_pgs(prs_id=pgs_id)
        p = run(["zcat", pgs_file, "|", 'head', '-n', "15"], shell=True)
        if p.returncode != 0:
            click.echo('The PGS file could not be viewed using cmd, please view it manually.')
        click.echo('Please check the PGS file viewed and provide us with the needed columns.')
    else:
        pgs_file = click.prompt('Please provide path to PGS file', type=str)
    id_col = click.prompt('Please provide the ID column number')
    allele = click.prompt('Please provide the effect allele column number')
    weight = click.prompt('Please provide the effect weight column number')
    cols = ' '.join([str(id_col), str(allele), str(weight)])
    file_type = click.prompt('Do you have a VCF file or binary files?', type=click.Choice(['vcf', 'bfile']))
    input_file = click.prompt('Please provide the path to input file', type=str)
    confirm = click.confirm('Please be aware that variant ID in both input file and pgs file need to match.'
                            'Do you want to continue?')
    if not confirm:
        return 'Ok. You still have the PGS file (if downloaded) but the scores were not calculated.'
    output_file = click.prompt('Please provide an output file path', type=str)
    if file_type == 'vcf':
        p = call(
            plink + " --vcf " + input_file + " --score " + pgs_file + " " + cols + " --out " + output_file,
            shell=True
        )
    else:
        p = call(
            plink + " --bfile " + input_file + " --score " + pgs_file + " " + cols + " --out " + output_file,
            shell=True
        )
    return 'Process is complete. Have a nice day!'


def download_pgs(
        *,
        prs_id,
):
    """
    Get PGS from pgscatalog.

    Parameters
    ----------
    prs_id : str
        the PRS ID in the pgscatalog.

    Returns
    -------
        a file containing the prs file

    """
    # make sure that the columns are present and matching
    resp = get('https://www.pgscatalog.org/rest/score/%s' % prs_id)
    prs_info = resp.json()
    if resp.status_code != 200 or not prs_info:
        raise Exception('The PRS ID might be wrong!')
    url = prs_info['ftp_scoring_file']
    prs_file = prs_id + '.gz'
    request.urlretrieve(url, prs_file)
    return prs_file
